import gym

# Implements the Frozen Lake game based on the gym library
# The goal of the game is to make it to the 'G' tile while avoiding 'H' tiles.
# Moves are defined as follows:
# LEFT = 0
# DOWN = 1
# RIGHT = 2
# RIGHT = 3

moves=["left", "down", "right", "up"]

class GymFrozenLake:
    def __init__(self, state=None, players=None):
        # Set up a gym environment for this game:
        self.env=gym.make('FrozenLake-v0', is_slippery=False)
        obs=self.env.reset()
        # This will be an array of tuples consisting of (move, cardsum)
        self.done=False
        self.reward=0
        # This is an array of strings:
        self.board=self.env.desc
        self.moveHist=["Start"]

        if state:
            self.state = state
        else:
            self.state = {
                "turn": 0,
                "players": players,
                "board": self.board
            }

    def print(self):
        print(self.moveHist)
        self.env.render()

    def validateMove(self, move):
        # Move should be a boolean (True or False depending on whether the player would like to continue)
        return self.env.action_space.contains(move)

    def makeMove(self, move):#assumes move has been validated
        print("makeMove %d" % move)
        observation, reward, done, info=self.env.step(move)
        # The board doesn't change
        self.done=done
        self.reward=reward
        self.moveHist.append(moves[move])
        print(self.reward)

    def postMove(self):
        # check for a winner:
        if self.done:
            if self.state['players'] is not None and self.reward>0:
                self.endGame(self.state['players'][0])
            else:
                self.endGame("Player lost.")
        else:
            self.state["turn"] = (self.state["turn"] + 1) % len(self.state["players"])

    def endGame(self, winner):
        self.state["Winner"] = winner
 



import copy

class Connect4:
    """This implements a game of Connect4"""
    def __init__(self, state=None, players=None, replayStates=None, moves=None):
        #Game Constants
        self.dims = (7 , 6)

        if state:
            self.state = state
        else:
            self.state = {
                "turn": 0,
                "board": [[0]*self.dims[1] for _ in range(self.dims[0])],
                "players":[]#an array of player ids
            }
            self.players = players
            for p in self.players:
                self.state["players"].append(p.id)
        if replayStates:
            self.replayStates = replayStates
        else:
            self.replayStates = []
        if moves:
            self.moves = moves
        else:
            self.moves = []

    def runGame(self, display=False):
        result = {
            "error":"none",
            "replayStates":self.replayStates,
            "moves":self.moves
        }


        self.replayStates.append(copy.deepcopy(self.state))
        while "Winner" not in self.state:
            # print("turn", self.game.state["turn"])
            player = self.players[self.state["turn"]]
            move = player.makeMove(self.state)
            self.moves.append(move)
            if not self.validateMove(move):
                result['rewards'] = self.genErrRewards(self.state["turn"])
                result['error'] = "AI : " + player.name + " made an illegal move"
                return result
            self.makeMove(move)
            self.replayStates.append(copy.deepcopy(self.state))
            self.postMove()
            if display:
                self.print()

        result['rewards'] = self.genRewards(self.state["Winner"])
        return result

    def genRewards(self, winner):
        rewards = []
        for i, player in enumerate(self.players):
            if i == winner:
                r = 1
            else:
                r = -1
            rewards.append((player.id, r))
        return rewards

    def genErrRewards(self, screwUp):
        rewards = []
        for i, player in enumerate(self.players):
            if i == screwUp:
                r = -1
            else:
                r = 0
            rewards.append((player.id, r))
        return rewards

    def print(self):
        print("---------------")
        for j in range(self.dims[1]):
            pStr = "|"
            for i in range(self.dims[0]):
                pStr += str(self.state["board"][i][self.dims[1] - 1 - j]) + "|"
            print(pStr)
        print("---------------")

    def validateMove(self, move):
        if move < 0 or move >= self.dims[0]:
            return False

        if self.state["board"][move][-1] == 0:
            return True

        return False

    def makeMove(self, move):#assumes move has been validated
        y = self.dims[1] - 1
        while self.state["board"][move][y] == 0:
            y += -1
            if y < 0:
                break

        marker = self.state["turn"] + 1
        self.state["board"][move][y+1] = marker




    def postMove(self):
        dirs = [(0,1),(1,0),(1,1),(-1,1)]
        #check for winner
        for i in range(self.dims[0]):
            j = 0
            while self.state["board"][i][j] != 0:
                match = self.state["board"][i][j]
                for dir in dirs:
                    streak = 1
                    dx = i + dir[0]
                    dy = j + dir[1]
                    if dx >= self.dims[0] or dy >= self.dims[1]:
                        continue
                    while self.state["board"][dx][dy] == match:
                        streak += 1
                        if streak == 4:
                            self.endGame(match - 1)
                        dx += dir[0]
                        dy += dir[1]

                        if dx >= self.dims[0] or dy >= self.dims[1]:
                            break

                j += 1
                if j == self.dims[1]:
                    break

        full = True
        for i in range(self.dims[1]):
            if self.state["board"][i][-1] == 0:
                full = False
        if full == True:
            self.endGame(-1)


        #if no winner incement move
        self.state["turn"] = (self.state["turn"] + 1) % 2



    def endGame(self, winner):
        #print("Winner", winner)
        #print(self.state["players"][winner])
        self.state["Winner"] = winner



from AIArena.AI import Player as Player

class Human_Connect4(Player):
    """This class is an example human interface for the Connect4 game. It allows
    for play over a terminal. 
    """
    def __init__(self):
        super().__init__("Human", "Connect4")

    def makeMove(self, state):
        """In order to support human play, this method reads a response from stdin and
        returns it.
        
        :return: Human move
        :rtype: int
        """
        move = "-1"
        while not move.isdigit() or int(move) >=7:
            move = input("Enter column number (0-6) to place piece: ")
        return int(move)



import AIArena.games
from AIArena.upload import upload
from AIArena.AI import AI as AI
from AIArena.Human import Human_Connect4 as Human_Connect4

name = "AIArena"

def runGame(Ais,gameName):
    result = {
        "error":"None"
    }
    #Game setup
    try:
        if gameName == "Connect4":
            game = AIArena.games.Connect4(players=Ais)
        elif gameName == "Arena":
            game = AIArena.games.Arena(players=Ais)
        else:
            result["error"] = "Game Creation: Bad Name"
            return result
    except:
        result["error"] = "Game Creation"
        return result

    #Game run
    #try:
    gameResult = game.runGame(False)
    result.update(gameResult)
    #except Exception as err:
    #    result["error"] = "Game Run"
    #    result["errMessage"] = err
    #    result["replayStates"] = game.replayStates
    #    return result
    #gameResult = game.runGame(False)
    #result["replayStates"] = game.replayStates
    result.update(gameResult)

    return result

def runMove(AI, state):
    try:
        move = AI.makeMove(state)
        results = {
            "error": "None"
        }
        if move == None:
            results['error'] = 'Bad move response'
        else:
            results['move'] = move
        return results
    except Exception as e:
        print("EXCEPTION:", e)
        return {'exception': e}





from AIArena.games.Connect4 import Connect4 as Connect4
from AIArena.games.GymBlackjack import GymBlackjack as GymBlackjack
from AIArena.games.GymFrozenLake import GymFrozenLake as GymFrozenLake
from AIArena.games.Arena import Arena as Arena
from AIArena.games.StockMarket import StockMarket as StockMarket



import gym

# Implements a simple blackjack game based on the gym library
# This is intended to be a *very* simple example to demonstrate how gym can be used
# to implement games.

class GymBlackjack:
    def __init__(self, state=None, players=None, playerClass=None):
        # Set up a gym environment for this game:
        self.env=gym.make('Blackjack-v0')
        self.env.reset()
        # This will be an array of tuples consisting of (move, cardsum)
        self.done=False
        self.reward=0
        # This tuple is formatted (playerhand, dealerhand, bool for whether player has a usable ace)
        self.board=self.env._get_obs()
        self.moveHist=[(None, (self.board[0], self.board[1]))]
        self.playerClass=None

        if state:
            self.state = state
        else:
            self.state = {
                "turn": 0,
                "players": players,
                "board": self.board
            }

    def print(self):
        print("---------------")
        print("Move history:")
        for elem in self.moveHist:
            if elem[0] is None:
                print("Initial. %s" % str(elem[1]))
            else:
                if elem[0]:
                    print("Hit. %s" % str(elem[1]))
                else:
                    print("Stick. %s" % str(elem[1]))
        print("---------------")

    def validateMove(self, move):
        # Move should be a boolean (True or False depending on whether the player would like to continue)
        return self.env.action_space.contains(move)

    def makeMove(self, move):#assumes move has been validated
        observation, reward, done, info=self.env.step(move)
        self.state['board']=observation
        self.done=done
        self.reward=reward
        self.moveHist.append((move, (observation[0], observation[1])))
        print(self.reward)

    def postMove(self):
        # check for a winner:
        if self.done:
            if self.state['players'] is not None and self.reward>0:
                self.endGame(self.state['players'][0])
            else:
                self.endGame("Dealer")
        else:
            self.state["turn"] = (self.state["turn"] + 1) % len(self.state["players"])

    def endGame(self, winner):
        self.state["Winner"] = winner
 



import requests
import json


class StockMarket:
    """This implements a game of a Stock Market Simulator"""
    def __init__(self, state=None, players=None, APIKey=None, StockData=None):
        #Game Constants
        self.startValue = 10**6
        self.interval = 15
        self.APIKey = APIKey
        if APIKey is None and StockData is None:
            with open("SM_SampleData.txt") as f:
                StockData = json.load(f)

        if players is None:
            self.players = []
        else:
            self.players = players

        if state:
            self.state = state
            for p in self.players:
                if p.id not in self.state["portfolios"]:#new players
                    self.state["portfolios"][p.id] = {
                        "cash":self.startValue,
                        "stocks":{},
                        "trades":[]
                    }
        else:
            self.state = {
                "endDate": "endDate",
                "symbols": ["FB","AMZN","AAPL","NFLX","GOOGL"],
                "data":{},
                "players": players,
                "portfolios":{}
            }

            for s in self.state["symbols"]:
                self.state["data"][s] = []

            for p in self.players:
                self.state["portfolios"][p.id] = {
                    "cash":self.startValue,
                    "stocks":{},
                    "trades":[]
                }

        self.cached = {}
        if StockData is not None:
            for s, data in StockData.items():
                self.cached[s] = True
                self.state["data"][s] = data
                #get last
                #last = sorted(list(data),reverse=True)[0]
                #self.state["data"][s]["last"] = data[last]




    def print(self):
        print("---------------")
        for id, port in self.state["portfolios"].items():
            print(id)
            print(port['cash'])
            print(port['stocks'])
        print("---------------")

    def gatherStockData(self):
        for sym in self.state["symbols"]:
            if sym in self.cached:
                continue
            prices = self.getStockData(sym, str(self.interval)+"min")
            self.state["data"][sym] = prices + self.state["data"][sym]
            #print(self.state["data"][sym])

    def getStockData(self, symbol, interval):
        payload = {
            "function": "TIME_SERIES_INTRADAY",
            "symbol": symbol,
            "interval": interval,
            "apikey": self.APIKey
        }

        r = requests.get("https://www.alphavantage.co/query", payload)
        if r.status_code == 200:
            data = json.loads(r.content.decode())
            #last = data["Meta Data"]["3. Last Refreshed"]
            prices = []
            for date, d in data["Time Series (" + interval + ")"].items():
                prices.append(float(d['4. close']))
            return prices
        else:
            print("An error occured:", r.status_code, r.content)

    def validateMove(self, move):
        pass


    def validateTrade(self, trade, portfolio):
        if trade["symbol"] in self.state["symbols"]:
            if trade["action"] == "buy":
                cost = trade["quantity"] * self.state["data"][trade["symbol"]][0]
                if cost < portfolio['cash']:
                    return True
            elif trade["action"] == "sell":
                if trade["symbol"] in portfolio["stocks"] and trade["quantity"] <= portfolio["stocks"][trade["symbol"]]:
                    return True
        return False

    def makeTrade(self, trade, portfolio, aid):#Assumes trade was validated
        cost = trade["quantity"] * self.state["data"][trade["symbol"]][0]
        if trade["action"] == "buy":
            self.state['portfolios'][aid]['cash'] -= cost
            if trade["symbol"] in portfolio['stocks']:
                self.state['portfolios'][aid]['stocks'][trade["symbol"]] += trade["quantity"]
            else:
                self.state['portfolios'][aid]['stocks'][trade["symbol"]] = trade["quantity"]
        elif trade["action"] == "sell":
            self.state['portfolios'][aid]['cash'] += cost
            self.state['portfolios'][aid]['stocks'][trade["symbol"]] -= trade["quantity"]

    def makeMove(self, move):#assumes move has been validated
        for trade in move['trades']:
            portfolio = self.state['portfolios'][move['aid']]
            if self.validateTrade(trade, portfolio):
                self.makeTrade(trade,portfolio,move['aid'])
                self.state['portfolios'][move['aid']]['trades'].append(trade)

    def postMove(self):
        pass

    def endGame(self, winner):
        #print("Winner", winner)
        #print(self.state["players"][winner])
        self.state["Winner"] = winner

    def exportData(self):
        data = {}
        for s in self.state["symbols"]:
            data[s] = self.state["data"][s]
        return data

    def exportValue(self, id):
        val = self.state['portfolios'][id]['cash']
        for ticker, quant in self.state['portfolios'][id]['stocks'].items():
            val += quant * self.state["data"][ticker][0]
        return val

import AIArena


if __name__ == "__main__":
    class lilStockBuddy(AIArena.AI):  # <2>
        def __init__(self, name):
            super().__init__(name, "StockMarket")

    ai1 = lilStockBuddy("lilStockBuddy")
    sm = StockMarket(players=[ai1])
    sm.gatherStockData()
    print(json.dumps(sm.exportData()))

    trades = [
        {'action':"buy",'symbol':"FB",'quantity':100}
    ]

    move = {
        'aid': ai1.id,
        'trades': trades
    }

    sm.print()
    sm.makeMove(move)
    sm.print()

    trades = [
        {'action': "buy", 'symbol': "FB", 'quantity': 100}
    ]

    sm = StockMarket(players=[ai1], state=sm.state)

    move = {
        'aid': ai1.id,
        'trades': trades
    }

    sm.makeMove(move)
    sm.print()
    print(sm.exportValue(ai1.id))
